/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers;

import de.bwaldvogel.liblinear.Feature;
import de.bwaldvogel.liblinear.FeatureNode;
import de.bwaldvogel.liblinear.Linear;
import de.bwaldvogel.liblinear.Model;
import de.bwaldvogel.liblinear.Parameter;
import de.bwaldvogel.liblinear.Problem;
import de.bwaldvogel.liblinear.SolverType;
import de.unibi.techfak.scie.classifiers.data.ClassificationResult;
import de.unibi.techfak.scie.classifiers.data.DataPoint;
import de.unibi.techfak.scie.classifiers.data.FeatureDictionary;
import de.unibi.techfak.scie.classifiers.data.FeatureMap;
import de.unibi.techfak.scie.classifiers.data.LabeledDataPoint;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Map.Entry;
import java.util.SortedMap;

/**
 * This is a wrapper for Benedikt Waldvogels liblinear implementation for Java.
 *
 * Please refer to http://liblinear.bwaldvogel.de/ for more information.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class LibLinearClassifier implements Classifier {

	private final SolverType solverType;
	private double C = 1;
	private double epsilon = 0.01;
	private Model trainedModel;
	private FeatureDictionary featureDictionary;
	private boolean silentCorrect = true;

	public LibLinearClassifier(SolverType solverType) {
		this.solverType = solverType;
	}

	public void setC(double C) {
		this.C = C;
	}

	public void setEpsilon(double epsilon) {
		this.epsilon = epsilon;
	}

	public double getC() {
		return C;
	}

	public double getEpsilon() {
		return epsilon;
	}

	public Model getTrainedModel() {
		return trainedModel;
	}

	@Override
	public FeatureDictionary getFeatureDictionary() {
		return featureDictionary;
	}

	/**
	 * LibLinear has a nasty habit of outputting the wrong confidence.
	 * This can be corrected and it is done, but usually silently. If you
	 * want to see every correction, use "setSilentCorrect(false)".
	 *
	 * @return true if the correction is done silently.
	 */
	public boolean isSilentCorrect() {
		return silentCorrect;
	}

	/**
	 * LibLinear has a nasty habit of outputting the wrong confidence.
	 * This can be corrected and it is done, but usually silently. If you
	 * want to see every correction, use "setSilentCorrect(false)".
	 *
	 * @param silentCorrect true if the correction shall be done silently.
	 */
	public void setSilentCorrect(boolean silentCorrect) {
		this.silentCorrect = silentCorrect;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public void train(ArrayList<LabeledDataPoint> trainingData) {
		//retrieve feature representation first.
		featureDictionary = new FeatureDictionary();
		final ArrayList<Feature[]> featureReps = new ArrayList<>();
		final ArrayList<Boolean> targets = new ArrayList<>();
		for (final LabeledDataPoint p : trainingData) {
			featureReps.add(
					transformToLibLinearFeatures(
							p.getData().getFeatureRepresentation(featureDictionary, true)));
			targets.add(p.getLabel());
		}
		//set up LibLinear problem description.
		final Problem problem = new Problem();
		problem.l = trainingData.size(); // number of training examples
		problem.n = featureDictionary.getNumberOfFeatures(); // number of features
		problem.x = new Feature[problem.l][]; // feature nodes
		problem.y = new double[problem.l]; // target values

		for (int p = 0; p < problem.l; p++) {
			if (targets.get(p)) {
				problem.y[p] = 1;
			} else {
				problem.y[p] = 0;
			}
			problem.x[p] = featureReps.get(p);
		}
		final Parameter parameter = new Parameter(solverType, C, epsilon);

		//train.
		this.trainedModel = Linear.train(problem, parameter);
		final double[] featureWeights = trainedModel.getFeatureWeights();
		for (int i = 0; i < problem.n; i++) {
			if (featureWeights[i] == 0) {
				featureDictionary.removeFeature(i + 1);
			}
		}
	}

	private static Feature[] transformToLibLinearFeatures(FeatureMap features) {
		final SortedMap<Integer, Double> featureValues = features.getFeatures();
		final Feature[] out = new Feature[featureValues.size()];
		int f = 0;
		for (final Entry<Integer, Double> featureValue : featureValues.entrySet()) {
			out[f] = new FeatureNode(
					featureValue.getKey(), featureValue.getValue());
			f++;
		}
		return out;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ClassificationResult classify(DataPoint data) {
		final Feature[] features = transformToLibLinearFeatures(data.
				getFeatureRepresentation(featureDictionary, false));
		final double[] confidences = new double[2];
		Linear.predictProbability(trainedModel, features, confidences);
		final boolean label = Linear.predict(trainedModel, features) > 0;
		/*
		 * There is some very strange behaviour of LibLinear in that it
		 * _sometimes_
		 * just decides to store the confidence for the correct label in
		 * confidences[0] and sometimes the other way around. I do not detect
		 * any system in there currently.
		 */
		double confidence = confidences[1];
		if (confidence > 0.5 != label) {
			if (!silentCorrect) {
				System.err.println("Warning! Liblinear seems to not behave in a sane way: Confidence=" + confidence + " but label=" + label + "\nSwapped the confidence (1-conf) to counteract that effect.");
			}
			confidence = 1 - confidence;
		}
		return new ClassificationResult(data, label, confidence);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public void writeParamters(Writer parameterOut,
			Writer featuresOut) throws IOException {
		this.trainedModel.save(parameterOut);
		this.featureDictionary.store(featuresOut);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public void readParameters(Reader parameterIn,
			Reader featuresIn) throws IOException {
		this.trainedModel = Model.load(parameterIn);
		this.featureDictionary = new FeatureDictionary();
		this.featureDictionary.load(featuresIn);
	}

}
