/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.classifiers;

import de.unibi.techfak.scie.classifiers.NonOverlappingCache.AnnotationCache;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.TreeMap;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

/**
 * This is a helper class to retrieve all annotations in a certain section of
 * the text.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class NonOverlappingCache extends HashMap<Class<? extends Annotation>, AnnotationCache> {

	public static int MAX_NUM_INSTANCES = 5;

	private static final ArrayList<JCas> cachedDocuments = new ArrayList<>();
	private static final ArrayList<NonOverlappingCache> cachedInstances = new ArrayList<>();

	private final JCas jcas;

	private NonOverlappingCache(JCas jcas) {
		this.jcas = jcas;
	}

	public static NonOverlappingCache getInstance(JCas jcas) {
		int cacheIdx = cachedDocuments.indexOf(jcas);
		final NonOverlappingCache instance;
		if (cacheIdx > -1) {
			/*
			 * If we have a cache for this document, move it to the end of the
			 * list to indicate that it was used recently.
			 */
			cachedDocuments.remove(cacheIdx);
			instance = cachedInstances.remove(cacheIdx);
			cachedDocuments.add(jcas);
			cachedInstances.add(instance);
		} else {
			/*
			 * Otherwise create a new cache.
			 */
			instance = new NonOverlappingCache(jcas);
			cachedDocuments.add(jcas);
			cachedInstances.add(instance);
			/*
			 * If we have now more caches than we want we remove the cache that
			 * was used longest ago.
			 */
			if (cachedDocuments.size() > MAX_NUM_INSTANCES) {
				cachedDocuments.remove(0);
				cachedInstances.remove(0);
			}
		}
		return instance;
	}

	public <X extends Annotation> ArrayList<X> selectBetween(Class<X> annoClass,
			int start, int end) {
		AnnotationCache<X> annoCache = get(annoClass);
		if (annoCache == null) {
			annoCache = new AnnotationCache<>();
			Collection<X> instances = JCasUtil.select(jcas, annoClass);
			for (final X instance : instances) {
				annoCache.put(instance.getBegin(), instance);
			}
			put(annoClass, annoCache);
		}
		/*
		 * get the correct sub map. We start one after the actual start and one
		 * before the actual and because the lowerKey and higherKey functions
		 * select
		 * strictly lower/strictly higher keys.
		 */
		Integer lower = annoCache.lowerKey(start + 1);
		if (lower == null) {
			lower = start;
		}
		Integer higher = annoCache.higherKey(end - 1);
		if (higher == null) {
			higher = end;
		}
		final ArrayList<X> out = new ArrayList<>(
				annoCache.subMap(lower, higher).values());
		return out;
	}

	/**
	 * This stores all instances of X in a document according to their starting
	 * position.
	 *
	 * @param <X> the annotation class for which all instances are stored.
	 */
	public static class AnnotationCache<X extends Annotation> extends TreeMap<Integer, X> {

		public AnnotationCache() {
		}

	}

}
