/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers;

import de.unibi.techfak.scie.classifiers.data.LabeledDataPoint;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Random;

/**
 * This is a wrapper class for some classifier training functionality.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielfeld.de
 */
public class TrainingUtils {

	/**
	 * Preprocesses the given data and enforces balanced classes by
	 * inserting data points of the under-represented label multiple
	 * times in the given list.
	 *
	 * @param data data points.
	 */
	public static void balanceDataSet(ArrayList<LabeledDataPoint> data) {
		int positives = 0;
		for (final LabeledDataPoint p : data) {
			if (p.getLabel()) {
				positives++;
			}
		}
		final int negatives = data.size() - positives;
		//sanity check: if negatives or positives are 0, we are unable to balance.
		if (positives > 0 && negatives > 0) {
			boolean balance = false;
			boolean underRepresentedLabel = false;
			int dist = negatives - positives;
			if (dist > 1) {
				balance = true;
				underRepresentedLabel = true;
			} else if (dist < -1) {
				balance = true;
				dist = -dist;
			}
			if (balance) {
				//find elements of the under-represented class.
				final ArrayList<LabeledDataPoint> underRepresented = new ArrayList<>();
				for (final LabeledDataPoint p : data) {
					if (p.getLabel() == underRepresentedLabel) {
						underRepresented.add(p);
					}
				}
				//copy elements of the under-represented class.
				final Random rand = new Random();
				for (int p = 0; p < dist; p++) {
					//insert a random member of the under-represented class ...
					int r = rand.nextInt(underRepresented.size());
					final LabeledDataPoint u = underRepresented.get(r);
					// ... at a random position.
					r = rand.nextInt(data.size() + 1);
					data.add(r, u);
				}
			}
		} else {
			System.err.println("Warning! The number of positives or negatives in"
					+ " the training data was 0. Therefore a balancing of the training"
					+ " set was not possible!");
		}
	}

	/**
	 * Chooses randomly testRatio datapoints for testing, trains the classifier
	 * with the rest and reports precision, recall and accuracy for the test
	 * set.
	 *
	 * @param trainingData the original training data. This list will be
	 * manipulated!
	 * If you want to keep your data, construct a copy!
	 * @param classifier the classifier that shall be trained.
	 * @param testRatio the ratio of test to training data (between 0 and 1). A
	 * value of about 0.1 to 0.2 is recommended.
	 * @param times defines how many times the training process should be
	 * repeated.
	 * @param comparator defines an optimality criterion to choose between
	 * repeated runs.
	 * @return The Classifier evaluation.
	 */
	public static ClassifierEvaluation simpleTraining(
			ArrayList<LabeledDataPoint> trainingData,
			final Classifier classifier, final double testRatio, int times,
			final Comparator<ClassifierEvaluation> comparator,
			boolean verbose) {
		Collections.sort(trainingData, new LabelComp());
		//Choose Test data.
		final int testNum = (int) Math.round((double) trainingData.size() * testRatio);
		final ArrayList<LabeledDataPoint> testData = new ArrayList<>();
		final Random rand = new Random();
		//check the number of positives in the training data.
		int posIdx = 0;
		while (trainingData.get(posIdx).getLabel()) {
			posIdx++;
		}
		for (int i = 0; i < testNum; i++) {
			final int r;
			if (posIdx == 1) {
				/*
				 * Ensure, that we take a negative example if only one
				 * positive is left ...
				 */
				r = posIdx + rand.nextInt(trainingData.size() - posIdx);
			} else if (trainingData.size() - posIdx == 1) {
				/*
				 * ... and ensure, that we take a positive example if only one
				 * negative is left.
				 */
				r = rand.nextInt(posIdx);
			} else {
				r = rand.nextInt(trainingData.size());
			}
			testData.add(trainingData.remove(r));
		}
		Tuple<String, String> optModel = null;
		ClassifierEvaluation optEval = null;
		for (int t = 0; t < times; t++) {
			//Training
			classifier.train(trainingData);
			final ClassifierEvaluation currentEval
					= new ClassifierEvaluation(classifier, trainingData, testData, verbose);
			if (optEval == null || comparator.compare(currentEval, optEval) == 1) {
				optEval = currentEval;
				optModel = storeModel(classifier);
			}
		}
		//after the sweep, read the optimal parameters into the classifier.
		loadModel(classifier, optModel);
		//and return the optimum evaluation.
		return optEval;
	}

	private static class LabelComp implements Comparator<LabeledDataPoint> {

		@Override
		public int compare(LabeledDataPoint o1, LabeledDataPoint o2) {
			return -Boolean.compare(o1.getLabel(), o2.getLabel());
		}

	}

	/**
	 * Trains the given classifier using crossvalidation and taking the result
	 * with the best evaluation result according to the given comparator.
	 *
	 * @param trainingData the original training data. This list will be
	 * manipulated!
	 * If you want to keep your data, construct a copy!
	 * @param classifier the classifier that shall be trained.
	 * @param folds the number of folds for a crossvalidation.
	 * @param times sets how often the training should be repeated within one
	 * fold.
	 * @param comparator a comparator for ClassifierEvaluation.
	 * @return the optimal evaluation.
	 */
	public static ClassifierEvaluation crossValidation(
			ArrayList<LabeledDataPoint> trainingData,
			final Classifier classifier, int folds,
			int times,
			final Comparator<ClassifierEvaluation> comparator,
			boolean verbose) {
		if (trainingData.isEmpty()) {
			throw new UnsupportedOperationException("No training data!");
		}
		//sort input.
		Collections.sort(trainingData, new LabelComp());
		//retrieve number of positives.
		int posIdx = 0;
		while (trainingData.get(posIdx).getLabel() && posIdx < trainingData.size()) {
			posIdx++;
		}
		if (posIdx < folds) {
			System.err.println("Warning! Data does not contain enough positive examples!"
					+ " Had to reduce the number of folds to " + posIdx);
			folds = posIdx;
		}
		if (trainingData.size() - posIdx < folds) {
			System.err.println("Warning! Data does not contain enough negative examples!"
					+ " Had to reduce the number of folds to " + (trainingData.size() - posIdx));
			folds = trainingData.size() - posIdx;
		}
		if (folds <= 1) {
			System.err.println("Warning! Number of folds is too small for cross validation! Using normal training instead!");
			classifier.train(trainingData);
			return new ClassifierEvaluation(classifier, trainingData, trainingData);
		}
		//partition the training data with the given number of folds.
		final ArrayList<ArrayList<LabeledDataPoint>> partitions = new ArrayList<>();
		final Random rand = new Random();
		final int foldSize = trainingData.size() / folds;
		for (int f = 0; f < folds - 1; f++) {
			final ArrayList<LabeledDataPoint> partition = new ArrayList<>();
			//add one positive and one negative example randomly.
			partition.add(trainingData.remove(rand.nextInt(posIdx)));
			posIdx--;
			partition.add(trainingData.remove(posIdx + rand.nextInt(trainingData.size() - posIdx)));
			for (int i = 0; i < foldSize - 1; i++) {
				//then choose randomly ...
				final int r;
				if (posIdx < folds - i) {
					/*
					 * ... but ensure, that we take a negative example if not
					 * many positives are left ...
					 */
					r = posIdx + rand.nextInt(trainingData.size() - posIdx);
				} else if (trainingData.size() - posIdx < folds - i) {
					/*
					 * ... and ensure, that we take a positive example if not
					 * many negatives are left.
					 */
					r = rand.nextInt(posIdx);
				} else {
					r = rand.nextInt(trainingData.size());
				}
				final LabeledDataPoint p = trainingData.remove(r);
				if (p.getLabel()) {
					posIdx--;
				}
				partition.add(p);
			}
			partitions.add(partition);
		}
		partitions.add(trainingData);
		//start the cross validation.
		Tuple<String, String> optModel = null;
		ClassifierEvaluation optEval = null;
		for (int f = 0; f < folds; f++) {
			//take one partition (f) for evaluation and all others (g) for training.
			final ArrayList<LabeledDataPoint> crossValTraining = new ArrayList<>();
			for (int g = 0; g < folds; g++) {
				if (f == g) {
					continue;
				}
				crossValTraining.addAll(partitions.get(g));
			}
			//train
			for (int t = 0; t < times; t++) {
				classifier.train(crossValTraining);
				//evaluate
				final ClassifierEvaluation currentEval
						= new ClassifierEvaluation(classifier, crossValTraining,
								partitions.get(f), verbose);
				//if we found a better result than before, remember those parameters.
				if (optEval == null || comparator.compare(currentEval, optEval) == 1) {
					optEval = currentEval;
					optModel = storeModel(classifier);
				}
			}
		}
		//after all cross validations, read the optimal parameters into the classifier.
		loadModel(classifier, optModel);
		//and return the optimum evaluation.
		return optEval;
	}

	/**
	 * Does the simple training but does the given parameter sweep in an outer
	 * loop.
	 *
	 * @param <C> the classifier class.
	 * @param trainingData the original training data.
	 * @param classifier the classifier that shall be trained.
	 * @param testRatio the ratio of test to training data (between 0 and 1). A
	 * value of about 0.1 to 0.2 is recommended.
	 * @param sweep the parameter sweep.
	 * @param times sets how many times a training run should be done for one
	 * setting.
	 * @param comparator a comparator for ClassifierEvaluation.
	 * @return the optimal evaluation.
	 */
	public static <C extends Classifier> ClassifierEvaluation simpleParameterSweep(
			final ArrayList<LabeledDataPoint> trainingData,
			final C classifier, double testRatio,
			final ParameterSweep<C> sweep,
			final int times,
			final Comparator<ClassifierEvaluation> comparator,
			boolean verbose) {
		Tuple<String, String> optModel = null;
		ClassifierEvaluation optEval = null;
		sweep.initializeParameter(classifier);
		while (sweep.hasNextParameter(classifier)) {
			//set next parameter.
			sweep.setNextParameter(classifier);
			//train & evaluate using simple training.
			final ClassifierEvaluation currentEval
					= simpleTraining(
							(ArrayList<LabeledDataPoint>) trainingData.clone(),
							classifier, testRatio, times, comparator,
							verbose);
			//if we found a better result than before, remember those parameters.
			if (optEval == null || comparator.compare(currentEval, optEval) == 1) {
				optEval = currentEval;
				optModel = storeModel(classifier);
				sweep.noteOptimal(classifier);
			}
		}
		//after the sweep, read the optimal parameters into the classifier.
		loadModel(classifier, optModel);
		//and return the optimum evaluation.
		return optEval;
	}

	/**
	 * Does the cross validation but does the given parameter sweep in an outer
	 * loop.
	 *
	 * @param <C> the classifier class.
	 * @param trainingData the original training data.
	 * @param classifier the classifier that shall be trained.
	 * @param folds the number of folds for a crossvalidation.
	 * @param times sets how often the training should be repeated within one
	 * fold.
	 * @param sweep the parameter sweep.
	 * @param comparator a comparator for ClassifierEvaluation.
	 * @return the optimal evaluation.
	 */
	public static <C extends Classifier> ClassifierEvaluation crossValidationSweep(
			final ArrayList<LabeledDataPoint> trainingData,
			final C classifier, int folds, int times,
			final ParameterSweep<C> sweep,
			final Comparator<ClassifierEvaluation> comparator,
			boolean verbose) {
		Tuple<String, String> optModel = null;
		ClassifierEvaluation optEval = null;
		sweep.initializeParameter(classifier);
		while (sweep.hasNextParameter(classifier)) {
			//set next parameter.
			sweep.setNextParameter(classifier);
			//train & evaluate using crossvalidation.
			final ClassifierEvaluation currentEval = crossValidation(
					(ArrayList<LabeledDataPoint>) trainingData.clone(),
					classifier, folds, times, comparator, verbose);
			//if we found a better result than before, remember those parameters.
			if (optEval == null || comparator.compare(currentEval, optEval) == 1) {
				optEval = currentEval;
				optModel = storeModel(classifier);
				sweep.noteOptimal(classifier);
			}
		}
		//after the sweep, read the optimal parameters into the classifier.
		loadModel(classifier, optModel);
		//and return the optimum evaluation.
		return optEval;
	}

	private static Tuple<String, String> storeModel(Classifier classifier) {
		try {
			final Tuple<String, String> out = new Tuple<>();
			final StringWriter parameterOut = new StringWriter();
			final StringWriter featuresOut = new StringWriter();
			classifier.writeParamters(parameterOut, featuresOut);
			parameterOut.close();
			featuresOut.close();
			out.setFirst(parameterOut.toString());
			out.setSecond(featuresOut.toString());
			return out;
		} catch (IOException ex) {
			throw new RuntimeException(ex);
		}
	}

	private static void loadModel(Classifier classifier, Tuple<String, String> data) {
		try {
			final StringReader parameterIn = new StringReader(data.getFirst());
			final StringReader featuresIn = new StringReader(data.getSecond());
			classifier.readParameters(parameterIn, featuresIn);
			parameterIn.close();
			featuresIn.close();
		} catch (IOException ex) {
			throw new RuntimeException(ex);
		}
	}

	/**
	 * This is an interface to define a parameter sweep on some classifier.
	 *
	 * @param <C> a classifier class for which this PaameterSweep is properly
	 * defined.
	 */
	public static interface ParameterSweep<C extends Classifier> {

		/**
		 * Initializes the parameter that shall be swept.
		 *
		 * @param classifier the classifier for which a parameter shall be
		 * swept.
		 */
		public void initializeParameter(C classifier);

		/**
		 * Sets the next parameter for the given classifier.
		 *
		 * @param classifier the classifier for which a parameter shall be
		 * swept.
		 */
		public void setNextParameter(C classifier);

		/**
		 * Returns true if the sweep for the given classifier is not yet
		 * finished.
		 *
		 * @param classifier the classifier for which a parameter shall be
		 * swept.
		 * @return true if the sweep for the given classifier is not yet
		 * finished.
		 */
		public boolean hasNextParameter(C classifier);

		/**
		 * This is called if the current value of the parameter is considered
		 * optimal.
		 *
		 * @param classifier the classifier for which a parameter shall be
		 * swept.
		 */
		public void noteOptimal(C classifier);
	}

}
