/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers.annotators;

import de.unibi.techfak.scie.descriptors.Aggregated;
import de.unibi.techfak.scie.descriptors.Annotation;
import de.unibi.techfak.scie.classifiers.Classifier;
import de.unibi.techfak.scie.classifiers.data.ClassificationResult;
import de.unibi.techfak.scie.classifiers.data.CoreDataPoint;
import de.unibi.techfak.scie.classifiers.data.CoreSlotCombinationDataPoint;
import de.unibi.techfak.scie.classifiers.data.RelationDataPoint;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Random;
import java.util.SortedMap;
import java.util.Stack;
import java.util.TreeMap;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.component.JCasAnnotator_ImplBase;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 * This is an abstract super class for annotators of relations according to the
 * following architecture:
 * <ol><li>There is a <em>core</em> or <em>trigger</em> element that needs to
 * be present for a relation to be annotated.</li>
 * <li>There are S <em>slot</em> elements that may be present.</li>
 * <li>The base set of viable candidates for core and slots is defined by an
 * annotation class (e.g. each element of slot 1 belongs to class A, each
 * element of slot 2 to class B, which might be even a superclass of A).</li>
 * <li>Which candidates from those sets should be aggregated to a relation can
 * be decided by S+2 different classifiers used iteratively:
 * <ol><li>First a core classifier decides whether a certain core instance is a
 * viable trigger for a relation at all.</li>
 * <li>Given that core element a classifier for each slot decides which
 * candidates are appropriate to be aggregated with that core element.</li>
 * <li>As candidate set for relation instances the cartesian product of the
 * candidate sets for each slot is constructed (also including the possibility
 * that a slot might not be filled at all).</li>
 * <li>The relation classifier classifiers each member of this
 * cartesian product and ranks it with a classifier confidence.</li>
 * <li>We then annotate at most the top K ranked relations on an overlapping
 * section of the text and at most the top R ranked relations in the document as
 * a whole. An implementing subclass can decide to run additional sanity
 * checks and thereby exclude relation instances from being annotated.</li>
 * </ol></li>
 * </ol>
 *
 * This already implements the basic annotation mechanism and only requires the
 * implementation of certain sub-routines of each implementing class.
 *
 * As additional heuristic to improve the runtime we provide users with the
 * option to limit candidates for each slot by a certain character distance.
 * This should be used with care! If distance is an important criterion to
 * classify whether slots are viable candidates the respective classifiers
 * should learn that regardless of a fixed constraint. However, often they do
 * not and the cartesian product mentioned above explodes. So for the sake of
 * runtime you might limit the cartesian product using that heuristic.
 *
 * Per default this is set to DEFAULT_DIST_LIMIT characters distance, which is a
 * very loose constraint only affecting outliers.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 * @param <C> the class of core elements.
 * @param <R> the relation class that is constructed by this annotator.
 */
public abstract class RelationAnnotator<C extends Annotation, R extends Aggregated> extends JCasAnnotator_ImplBase {

	private final Class<C> coreClass;
	private final Classifier coreClassifier;
	private final ArrayList<SlotSpecification> slotSpecifications = new ArrayList<>();
	private final Classifier relationClassifier;
	public static final int DEFAULT_DIST_LIMIT = 3000;
	private int distLimit = DEFAULT_DIST_LIMIT;
	public static final int DEFAULT_K = 3;
	private int K = DEFAULT_K;
	public static final int DEFAULT_R = 6;
	private int R = DEFAULT_R;

	public RelationAnnotator(Class<C> coreClass, Classifier coreClassifier,
			Classifier relationClassifier) {
		this.coreClass = coreClass;
		this.coreClassifier = coreClassifier;
		this.relationClassifier = relationClassifier;
	}

	public Class<C> getCoreClass() {
		return coreClass;
	}

	public Classifier getCoreClassifier() {
		return coreClassifier;
	}

	public ArrayList<SlotSpecification> getSlotSpecifications() {
		return slotSpecifications;
	}

	public Classifier getRelationClassifier() {
		return relationClassifier;
	}

	/**
	 * Returns the limit (in characters) a core instance may have to a
	 * slot instance.
	 *
	 * As additional heuristic to improve the runtime we provide users with the
	 * option to limit candidates for each slot by a certain character distance.
	 * This should be used with care! If distance is an important criterion to
	 * classify whether slots are viable candidates the respective classifiers
	 * should learn that regardless of a fixed constraint. However, often they
	 * do not and the cartesian product mentioned above explodes. So for the
	 * sake of runtime you might limit the cartesian product using that
	 * heuristic.
	 *
	 * Per default this is set to DEFAULT_DIST_LIMIT characters distance, which
	 * is a very loose constraint only affecting outliers.
	 *
	 * @return the limit (in characters) a core instance may have to a
	 * slot instance.
	 */
	public int getDistLimit() {
		return distLimit;
	}

	/**
	 * Sets the limit (in characters) a core instance may have to a
	 * slot instance.
	 *
	 * As additional heuristic to improve the runtime we provide users with the
	 * option to limit candidates for each slot by a certain character distance.
	 * This should be used with care! If distance is an important criterion to
	 * classify whether slots are viable candidates the respective classifiers
	 * should learn that regardless of a fixed constraint. However, often they
	 * do not and the cartesian product mentioned above explodes. So for the
	 * sake of runtime you might limit the cartesian product using that
	 * heuristic.
	 *
	 * Per default this is set to DEFAULT_DIST_LIMIT characters distance, which
	 * is a very loose constraint only affecting outliers.
	 *
	 * @param distLimit the limit (in characters) a core instance may have to a
	 * slot instance.
	 */
	public void setDistLimit(int distLimit) {
		this.distLimit = distLimit;
	}

	/**
	 * Returns the number of annotations that may be present on an
	 * overlapping piece of text. If the number of such annotations gets higher
	 * than K somewhere in the text, we only take the K ones with highest
	 * confidence. If K is higher than R it has no effect.
	 *
	 * @return the number of annotations that may be present on an
	 * overlapping piece of text.
	 */
	public int getK() {
		return K;
	}

	/**
	 * Returns the number of annotations that may be present on an
	 * overlapping piece of text. If the number of such annotations gets higher
	 * than K somewhere in the text, we only take the K ones with highest
	 * confidence. If K is higher than R it has no effect.
	 *
	 *
	 * @param K the number of annotations that may be present on an
	 * overlapping piece of text.
	 */
	public void setK(int K) {
		this.K = K;
	}

	/**
	 * Returns the number of annotations that may be present in the text. If the
	 * number of annotations gets higher than R we only take the R ones with
	 * highest confidence.
	 *
	 * @return the number of annotations that may be present in the text.
	 */
	public int getR() {
		return R;
	}

	/**
	 * Sets the number of annotations that may be present in the text. If the
	 * number of annotations gets higher than R we only take the R ones with
	 * highest confidence.
	 *
	 * @param R the number of annotations that may be present in the text.
	 */
	public void setR(int R) {
		this.R = R;
	}

	/**
	 * Annotates a given document in its JCas form. Please refer to the
	 * documentation of the RelationAnnotator class for more information.
	 *
	 * @param jcas a document given as a JCas instance.
	 * @throws AnalysisEngineProcessException
	 */
	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final int S = slotSpecifications.size();
		//store the created relations in data structures to allow for
		//controlling the K and R limit.
		final PositionMap posMap = new PositionMap();
		final ConfidenceMap confidenceMap = new ConfidenceMap();
		for (final C coreAnno : JCasUtil.select(jcas, coreClass)) {
			final ClassificationResult coreClassification
					= coreClassifier.classify(
							createCoreDataPoint(coreAnno, jcas));
			//we do not use cores that are not accepted by the core classifier.
			if (!coreClassification.getLabel()) {
				continue;
			}

			//create the set of viable candidates for each slot given the core.
			final ArrayList<ArrayList<SlotCandidate<? extends Annotation>>> slotCandidatess = new ArrayList<>();
			int productSize = 1;
			for (int s = 0; s < S; s++) {
				final SlotSpecification spec = slotSpecifications.get(s);
				final ArrayList<SlotCandidate<? extends Annotation>> slotCandidates
						= getSlotCandidates(spec, coreAnno, jcas);
				productSize *= slotCandidates.size();
				slotCandidatess.add(slotCandidates);
			}
			if (productSize > 300) {
				System.out.println("Warning! The cartesian product for core "
						+ coreAnno.getCoveredText() + " is very large ("
						+ productSize + " potential relations). Consider "
						+ "lowering the distance limit!");
			}
			/*
			 * Now construct the cartesian product itself.
			 */
			final Stack<LinkedList<Integer>> cartesianStack = new Stack<>();
			cartesianStack.push(new LinkedList<Integer>());
			while (!cartesianStack.empty()) {
				final LinkedList<Integer> currentProduct = cartesianStack.pop();
				if (currentProduct.size() == S) {
					/*
					 * if we have accumulated one candidate for each
					 * slot, we are finished
					 */
					final SlotCandidate[] slots = new SlotCandidate[S];
					//construct slot arrays
					for (int s = 0; s < S; s++) {
						final int candidateIdx = currentProduct.get(s);
						final ArrayList<SlotCandidate<? extends Annotation>> slotCandidates
								= slotCandidatess.get(s);
						slots[s] = slotCandidates.get(candidateIdx);
					}
					//now construct the datapoint.
					final RelationDataPoint<C> relDataPoint
							= createDataPoint(coreAnno, coreClassification,
									slots, jcas);
					//classify it as well.
					final ClassificationResult relationResult
							= relationClassifier.classify(relDataPoint);
					//check if its confidence is high enough to be annotated
					//even given the R limit.
					if (!limitNumberOfAnnotations(posMap, confidenceMap,
							relationResult.getConfidence())) {
						continue;
					}
					final R relation = createRelationObject(
							coreAnno, coreClassification, slots, jcas);
					if (relation == null) {
						//if the returned object is null, sanity checks
						//failed and we do not annotate that relation.
						continue;
					}
					int begin = coreAnno.getBegin();
					int end = coreAnno.getEnd();
					for (final SlotCandidate slotCandidate : slots) {
						if (slotCandidate == null) {
							continue;
						}
						if (slotCandidate.getAnnotation().getBegin() < begin) {
							begin = slotCandidate.getAnnotation().getBegin();
						}
						if (slotCandidate.getAnnotation().getEnd() > end) {
							end = slotCandidate.getAnnotation().getEnd();
						}
					}
					if (!limitOverlapping(posMap, confidenceMap, begin, end,
							relationResult.getConfidence())) {
						continue;
					}

					relation.setBegin(begin);
					relation.setEnd(end);
					relation.setProbability(relationResult.getConfidence());
					relation.addToIndexes();
					//store the new relation in our data structures.
					posMap.add(relation);
					confidenceMap.add(relation);
					continue;
				}
				/*
				 * If we are not finished yet we iterate through all candidates
				 * of the current slot and copy a new product instance for all
				 * of them.
				 */
				final ArrayList<SlotCandidate<? extends Annotation>> slotCandidates
						= slotCandidatess.get(currentProduct.size());
				boolean needsCopy = false;
				for (int c = 0; c < slotCandidates.size(); c++) {
					final LinkedList<Integer> newProduct;
					if (!needsCopy) {
						newProduct = currentProduct;
						needsCopy = true;
					} else {
						newProduct = (LinkedList<Integer>) currentProduct.clone();
						newProduct.pollLast();
					}
					newProduct.add(c);
					cartesianStack.push(newProduct);
				}
			}
		}
	}

	private <S extends Annotation> ArrayList<SlotCandidate<S>> getSlotCandidates(
			SlotSpecification<S> spec, C coreAnno, JCas jcas) {
		final ArrayList<SlotCandidate<S>> slotCandidates = new ArrayList<>();
		//the empty canidate is always possible.
		slotCandidates.add(null);
		final int cStart = coreAnno.getBegin();
		final int cEnd = coreAnno.getEnd();
		for (final S slotAnno : JCasUtil.select(jcas, spec.getSlotClass())) {
			/*
			 * We do not want to consider annotations that are further away than
			 * the distance limit.
			 *
			 * This is not a nice heuristic, because in principle the
			 * classifiers should do this automatically. But practical
			 * experience has shown that they do not do so necessarily.
			 */
			final int dist;
			if (slotAnno.getEnd() < cStart) {
				dist = cStart - slotAnno.getEnd();
			} else if (slotAnno.getBegin() > cEnd) {
				dist = slotAnno.getBegin() - cEnd;
			} else {
				dist = 0;
			}
			if (dist > distLimit) {
				continue;
			}
			final CoreSlotCombinationDataPoint<C, ? extends Annotation> slotComb
					= createCoreSlotCombinationDataPoint(
							spec, coreAnno, slotAnno, jcas);
			final ClassificationResult slotClassification
					= spec.getSlotClassifier().classify(slotComb);
			/*
			 * we do not want to use slot fillers with negative
			 * classification results.
			 */
			if (slotClassification.getLabel()) {
				slotCandidates.add(
						new SlotCandidate<>(slotAnno, slotClassification));
			}
		}
		return slotCandidates;
	}

	/**
	 *
	 * @return true if the overlapping limit allows for the new annotation.
	 */
	private boolean limitOverlapping(PositionMap posMap, ConfidenceMap confMap,
			int newBegin, int newEnd, double newConfidence) {
		//if there is no need for overlapping checks we return true.
		if (posMap.isEmpty() || this.K > this.R) {
			return true;
		}
		/*
		 * construct the set of all current annotations overlapping with the
		 * new one, namely: All annotations that start before the new one ends
		 * and end after the new one begins.
		 */
		final HashSet<R> overlappings = posMap.getOverlappings(newBegin, newEnd);
		/*
		 * Now check if that would contain more than K annotations if the new
		 * one is added.
		 */
		if (overlappings.size() >= this.K) {
			//if so we remove the annotations with least confidence.
			final ArrayList<R> sortedOverlappings = new ArrayList<>(overlappings);
			Collections.sort(sortedOverlappings, confidenceComp);
			double currentConfidence = 0;
			int limit = this.K - 1;
			while (sortedOverlappings.size() > limit) {
				final R lowestRel = sortedOverlappings.get(0);
				currentConfidence = lowestRel.getProbability();
				//if the current confidence is better (or equal) to our new one, we do
				//not add the new annotation.
				if (currentConfidence >= newConfidence) {
					limit++;
				} else {
					//otherwise we remove the old one.
					lowestRel.removeFromIndexes();
					sortedOverlappings.remove(0);
					posMap.remove(lowestRel);
					confMap.remove(lowestRel);
				}
			}
			//we add the new annotation if it is better than everything we removed.
			return newConfidence > currentConfidence;
		} else {
			return true;
		}
	}

	/**
	 *
	 * @return true if the limit of annotations in the whole document
	 */
	private boolean limitNumberOfAnnotations(PositionMap posMap,
			final ConfidenceMap confidenceMap, final double newConfidence) {
		//if the number of created annotations is small enough, return true.
		if (confidenceMap.size() < this.R) {
			return true;
		}
		//otherwise we have to remove the least confident annotations until
		//our limit is met again.
		int limit = this.R - 1;
		double currentConfidence = 0;
		while (confidenceMap.size() > limit) {
			currentConfidence = confidenceMap.firstKey();
			//if our new confidence is not better than the current one, we do
			//not add the new one.
			if (currentConfidence >= newConfidence) {
				limit++;
			} else {
				//otherwise we remove an old one.
				final R removed = confidenceMap.removeRandomly(currentConfidence);
				if (removed == null) {
					throw new UnsupportedOperationException("Internal error! Relation could not be removed from confidence map!");
				}
				posMap.remove(removed);
				removed.removeFromIndexes();
			}
		}
		return newConfidence > currentConfidence;
	}

	private final class PositionMap extends TreeMap<Integer, TreeMap<Integer, LinkedList<R>>> {

		public PositionMap() {
		}

		public void add(R relation) {
			TreeMap<Integer, LinkedList<R>> endMap
					= this.get(relation.getBegin());
			if (endMap == null) {
				endMap = new TreeMap<>();
				this.put(relation.getBegin(), endMap);
			}
			LinkedList<R> endList = endMap.get(relation.getEnd());
			if (endList == null) {
				endList = new LinkedList<>();
				endMap.put(relation.getEnd(), endList);
			}
			endList.add(relation);
		}

		public void remove(R relation) {
			TreeMap<Integer, LinkedList<R>> endMap
					= this.get(relation.getBegin());
			if (endMap == null) {
				return;
			}
			LinkedList<R> endList = endMap.get(relation.getEnd());
			if (endList == null) {
				return;
			}
			endList.remove(relation);
		}

		public HashSet<R> getOverlappings(int begin, int end) {
			final HashSet<R> overlappings = new HashSet<>();
			final SortedMap<Integer, TreeMap<Integer, LinkedList<R>>> headMap
					= this.headMap(end);
			for (final TreeMap<Integer, LinkedList<R>> endMap : headMap.values()) {
				for (final LinkedList<R> overlappingPositionList : endMap.tailMap(begin).values()) {
					overlappings.addAll(overlappingPositionList);
				}
			}
			return overlappings;
		}
	}

	private final class ConfidenceMap extends TreeMap<Double, LinkedList<R>> {

		private int size = 0;

		public ConfidenceMap() {
		}

		@Override
		public int size() {
			return size;
		}

		public void add(R relation) {
			LinkedList<R> confidenceList
					= this.get(relation.getProbability());
			if (confidenceList == null) {
				confidenceList = new LinkedList<>();
				this.put(
						relation.getProbability(), confidenceList);
			}
			confidenceList.add(relation);
			size++;
		}

		public void remove(R relation) {
			final LinkedList<R> confidenceList
					= this.get(relation.getProbability());
			if (confidenceList == null) {
				return;
			}
			if (confidenceList.remove(relation)) {
				size--;
			}
			if (confidenceList.isEmpty()) {
				this.remove(relation.getProbability());
			}
		}

		public R removeRandomly(final double confidence) {
			final LinkedList<R> confidenceList = this.get(confidence);
			if (confidenceList == null) {
				return null;
			}
			if (confidenceList.size() == 1) {
				size--;
				this.remove(confidence);
				return confidenceList.get(0);
			} else {
				final Random rand = new Random();
				final int r = rand.nextInt(confidenceList.size());
				size--;
				return confidenceList.remove(r);
			}
		}

	}

	private final Comparator<R> confidenceComp = new Comparator<R>() {

		@Override
		public int compare(R t, R t1) {
			return Double.compare(t.getProbability(), t1.getProbability());
		}

	};

	/**
	 * This method should create a RelationDataPoint given a core instance,
	 * a core classification and Candidates for all slots (note that the
	 * candidates might be null).
	 *
	 * @param core a core instance.
	 * @param coreClassification the classification of that core instance
	 * including the classification confidence (the label will be true).
	 * @param slotCandidates the candidates for each slot. Note that each entry
	 * of that array might be null if the slot is not filled.
	 * @param jcas the current JCas object.
	 *
	 * @return a RelationDataPoint containing the given information.
	 */
	public abstract RelationDataPoint<C> createDataPoint(C core,
			ClassificationResult coreClassification,
			SlotCandidate[] slotCandidates,
			JCas jcas);

	/**
	 * This method should create a CoreDataPoint given a core instance.
	 *
	 * @param core a core instance.
	 * @param jcas the current JCas object.
	 *
	 * @return a CoreDataPoint containing the given information.
	 */
	public abstract CoreDataPoint<C> createCoreDataPoint(C core, JCas jcas);

	/**
	 * This method should create a CoreSlotCombination given a
	 * SlotSpecification, a core instance and a Slot instance.
	 *
	 * @param spec the current SlotSpecification.
	 * @param code the current Core instance.
	 * @param slot the current slot instance.
	 * @param jcas the current JCas object.
	 *
	 * @return a CoreSlotCombination containing the given data.
	 */
	public abstract CoreSlotCombinationDataPoint<C, ? extends Annotation> createCoreSlotCombinationDataPoint(
			final SlotSpecification spec, C code, Annotation slot, JCas jcas);

	/**
	 * This is only needed to create an empty relation object and to add
	 * relation-specific information. Any general information (that is used for
	 * all relations) will be added in the RelationAnnotator parent class. That
	 * includes:
	 * <ul><li>begin and end index</li>
	 * <li>confidence</li></ul>
	 *
	 * The implementing method is allowed to apply sanity checks and return 0
	 * if those checks are not met. In that case the relation will not be
	 * annotated.
	 *
	 * @param core the core instance.
	 * @param coreClassification the ClassificationResult for the core.
	 * @param slots all SlotCandidates (some of them may be null).
	 * @param jcas the current JCas object.
	 *
	 * @return a newly constructed relation object or null if sanity checks
	 * are not met.
	 */
	public abstract R createRelationObject(C core, ClassificationResult coreClassification,
			SlotCandidate[] slots, JCas jcas);
}
