/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.classifiers.data;

import de.unibi.techfak.scie.classifiers.FeatureUtils;
import de.unibi.techfak.scie.descriptors.Annotation;
import java.util.Collection;
import java.util.HashSet;
import java.util.Objects;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 * This models the core of some relation with its relevant features.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 * @param <C> the class of this core.
 */
public abstract class CoreDataPoint<C extends Annotation> implements DataPoint {

	private final C core;
	private final JCas jcas;

	public CoreDataPoint(C core, JCas jcas) {
		this.core = core;
		this.jcas = jcas;
	}

	public C getCore() {
		return core;
	}

	public JCas getJcas() {
		return jcas;
	}

	/**
	 * Returns a FeatureMap representing this CoreDataPoint sparsely. This
	 * adds some standard features to the mix:
	 *
	 * <ul><li>Relative Start index in the text</li>
	 * <li>Relative End index in the text</li>
	 * <li>Relative annotation length</li>
	 * <li>dictionary confidence; if this is not a dictionary-based word, a 1
	 * is used.</li>
	 * <li>word identity feature (boolean feature)</li>
	 * <li>POS tags of core word (boolean features)</li>
	 * <li>dictionary id identity (boolean feature); if these are not
	 * dictionary-based words, a 0 is used.</li></ul>
	 *
	 * Additional features are added by the addSpecialFeatures method depending
	 * on the specific implementation.
	 *
	 * @return a FeatureMap representing this CoreDataPoint sparsely.
	 */
	@Override
	public FeatureMap getFeatureRepresentation(FeatureDictionary dict,
		boolean addFeaturesToDictionary) {
		final FeatureMap features = new FeatureMap(dict);
		//text position features.
		features.addNumericFeature("coreBegin", FeatureUtils.getRelativeBegin(core, jcas),
			addFeaturesToDictionary);
		features.addNumericFeature("coreEnd", FeatureUtils.getRelativeEnd(core, jcas),
			addFeaturesToDictionary);
		//length feature
		features.addNumericFeature("coreLength", FeatureUtils.getRelativeLength(core, jcas),
			addFeaturesToDictionary);
		//get confidence feature
		features.addNumericFeature("coreConfidence", FeatureUtils.getConfidence(core),
			addFeaturesToDictionary);
		//word identity feature
		features.addBooleanFeature(FeatureUtils.cleanWord(core.getCoveredText()),
			addFeaturesToDictionary);
		//POS tag features.
		for (final String pos : FeatureUtils.getPOSTags(core, jcas)) {
			features.addBooleanFeature("POS" + pos, addFeaturesToDictionary);
		}
		//dictionary id identity feature
		features.addBooleanFeature("coreID" + FeatureUtils.getDictionaryID(core),
			addFeaturesToDictionary);
		//add special features.
		addSpecialFeatures(features, addFeaturesToDictionary);
		return features;
	}

	/**
	 * This is to be implemented by a subclass to add domain-specific features
	 * to the mix.
	 *
	 * @param stdFeatures featureMap already filled with standard features.
	 * @param addFeaturesToDictionary this should be set to true during
	 * training, because then unknown features should be added to the
	 * FeatureDictionary. Outside training this should be set to false because
	 * then we want to treat every unknown feature as zero implicitly.
	 */
	public abstract void addSpecialFeatures(FeatureMap stdFeatures,
		boolean addFeaturesToDictionary);

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 41 * hash + Objects.hashCode(this.core);
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final CoreDataPoint<?> other = (CoreDataPoint<?>) obj;
		if (!Objects.equals(this.core, other.core)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "{" + core.getCoveredText() + "}";
	}

	/**
	 * The CoreTrainingDataReader is pretty straightforward and just creates one
	 * DataPoint for each instance of the core class. If this core instance is
	 * specified in a RawRelation it will be treated as a positive example,
	 * otherwise as a negative. Now restrictions to the search space are
	 * applied, such that the full space of possible training data is used.
	 *
	 * @param <C> the core class.
	 */
	public static abstract class CoreTrainingDataReader<C extends Annotation>
		implements TrainingDataReader {

		private final Class<C> coreClass;
		private final String rawRelationClass;

		public CoreTrainingDataReader(Class<C> coreClass, String rawRelationClass) {
			this.coreClass = coreClass;
			this.rawRelationClass = rawRelationClass;
		}

		public Class<C> getCoreClass() {
			return coreClass;
		}

		public String getRawRelationClass() {
			return rawRelationClass;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void readTrainingData(JCas jcas,
			Collection<RawRelation> rawRelationInstances,
			Collection<LabeledDataPoint> trainingDataCollection) {
			final Collection<C> coreInstances = JCasUtil.select(jcas, coreClass);
			final HashSet<Long> positiveData = new HashSet<>();
			// get relevant IDs
			for (final RawRelation relationInstance : rawRelationInstances) {
				if (!relationInstance.relationClass.equals(rawRelationClass)) {
					continue;
				}
				for (final Long id : relationInstance.ids) {
					positiveData.add(id);
				}
			}
			//iterate over instances and create training data points out of them.
			for (final C coreInstance : coreInstances) {
				final CoreDataPoint<C> dataPoint = createDataPoint(coreInstance, jcas);
				trainingDataCollection.add(
					new LabeledDataPoint(
						positiveData.contains(coreInstance.getAnnotationId()),
						dataPoint));
			}
		}

		public abstract CoreDataPoint<C> createDataPoint(C core, JCas jcas);

	}
}
