/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

/**
 * This represents relations for training purposes. Please note that a
 * RawRelation is not equivalent to a RelationDataPoint, because a RawRelation
 * is allowed to specify multiple members for each slot. The RelationDataPoints
 * that can be constructed using this RawRelation are implicitly given as
 * follows:
 *
 * Say the relationClass of this RawRelation specifies 2 slots and 1 core.
 *
 * This specific RawRelation now contains IDs for 2 core candidates, 3
 * candidates for the first slot and 1 candidate for the second slot. Let C, S_1
 * and S_2 be the respective sets. Than the RelationDataPoints can be
 * constructed using the Cartesian Product C x S_1 x S_2, leading to 2*3*1 = 6
 * RelationDataPoints.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class RawRelation {

    public final String relationClass;
    public final ArrayList<Long> ids = new ArrayList<>();

    public RawRelation(String relationClass) {
        this.relationClass = relationClass;
    }

    /**
     * This reads all RawRelations from the given .rel file.
     *
     * @param relationFile a .rel file.
     * @return a list of RawRelations contained in the given .rel file.
     * @throws IOException is thrown is something during the input file reading
     * goes wrong.
     */
    public static ArrayList<RawRelation> readRawRelations(final File relationFile) throws IOException {
        final ArrayList<RawRelation> out = new ArrayList<>();
        final BufferedReader reader = new BufferedReader(new FileReader(relationFile));
        String line;
        while ((line = reader.readLine()) != null) {
            final String[] split = line.split("\\:");
            final String relationClass = split[0];
            final RawRelation rel = new RawRelation(relationClass);
            final String[] annotationIdStrings = split[1].split("\\,");
            for (final String idString : annotationIdStrings) {
                rel.ids.add(Long.parseLong(idString));
            }
            out.add(rel);
        }
        reader.close();
        return out;
    }

}
