/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.classifiers.data;

import de.unibi.techfak.scie.classifiers.Classifier;
import de.unibi.techfak.scie.classifiers.FeatureUtils;
import de.unibi.techfak.scie.descriptors.Annotation;
import de.unibi.techfak.scie.descriptors.Token;
import de.unibi.techfak.scie.classifiers.NonOverlappingCache;
import de.unibi.techfak.scie.classifiers.annotators.SlotCandidate;
import de.unibi.techfak.scie.classifiers.annotators.SlotSpecification;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Random;
import java.util.Stack;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 * This defines a relation instance with one specific instance for the core and
 * each slot. Note that our architecture specifies that the core and every slot
 * filler have been previously classified by a separate classifier. Therefore we
 * also store the respective ClassificationResult inside this DataPoint.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 * @param <C> Core class
 */
public abstract class RelationDataPoint<C extends Annotation> implements DataPoint {
	
	private final C core;
	private final ClassificationResult coreClassification;
	private final SlotCandidate[] slots;
	private final JCas jcas;
	
	public RelationDataPoint(C core, ClassificationResult coreClassification,
			SlotCandidate[] slots, JCas jcas) {
		this.core = core;
		this.coreClassification = coreClassification;
		this.slots = slots;
		this.jcas = jcas;
	}
	
	public C getCore() {
		return core;
	}
	
	public ClassificationResult getCoreClassification() {
		return coreClassification;
	}
	
	public SlotCandidate[] getSlots() {
		return slots;
	}
	
	public JCas getJcas() {
		return jcas;
	}

	/**
	 * Returns a FeatureMap representing this RelationDataPoint sparsely. This
	 * adds some standard features to the mix:
	 *
	 * <ul><li>Relative relation begin.</li>
	 * <li>Relative relation end.</li>
	 * <li>Relative relation length.</li>
	 * <li>Core classifier confidence.</li>
	 * <li>slot classifier confidences.</li></ul>
	 *
	 * Additional features are added by the addSpecialFeatures method depending
	 * on the specific implementation.
	 *
	 * @return a FeatureMap representing this RelationDataPoint sparsely.
	 */
	@Override
	public FeatureMap getFeatureRepresentation(FeatureDictionary dict,
			boolean addFeaturesToDictionary) {
		final FeatureMap features = new FeatureMap(dict);
		//text span features.
		int min = core.getBegin();
		int max = core.getEnd();
		for (final SlotCandidate slotCandidate : slots) {
			if (slotCandidate == null) {
				continue;
			}
			final Annotation slot = slotCandidate.getAnnotation();
			if (slot.getBegin() < min) {
				min = slot.getBegin();
			}
			if (slot.getEnd() > max) {
				max = slot.getEnd();
			}
		}
		features.addNumericFeature("spanBegin",
				FeatureUtils.getRelativeIndex(min, jcas),
				addFeaturesToDictionary);
		features.addNumericFeature("spanEnd",
				FeatureUtils.getRelativeIndex(max, jcas),
				addFeaturesToDictionary);
		
		features.addNumericFeature("spanSize",
				FeatureUtils.getRelativeIndex(max - min, jcas),
				addFeaturesToDictionary);

		//number of tokens in the span.
		final List<Token> tokens = NonOverlappingCache.getInstance(jcas).
				selectBetween(Token.class, min, max);
		features.addNumericFeature("spanTokenNum", tokens.size(),
				addFeaturesToDictionary);
		//binned number of tokens.
		if (tokens.size() > 50) {
			features.addBooleanFeature("moreThan" + 50 + "SpanTokens",
					addFeaturesToDictionary);
		} else {
			int startT = tokens.size() + 5 - (tokens.size() % 5);
			for (int t = startT; t < 50; t += 5) {
				if (tokens.size() < t) {
					features.addBooleanFeature("lessThan" + t + "SpanTokens",
							addFeaturesToDictionary);
				}
			}
		}
		//token word identity features.
		for (final Token t : tokens) {
			features.addBooleanFeature("inBetween" + t.getCoveredText(),
					addFeaturesToDictionary);
		}

		//confidence features
		features.addNumericFeature("coreConfidence",
				coreClassification.getConfidence(),
				addFeaturesToDictionary);
		for (int s = 0; s < slots.length; s++) {
			if (slots[s] != null) {
				features.addNumericFeature("confidenceSlot" + s,
						slots[s].getResult().getConfidence(),
						addFeaturesToDictionary);
			}
		}
		//boolean feature that return if the slot is filled at all.
		int filledSlots = 0;
		for (int s = 0; s < slots.length; s++) {
			if (slots[s] != null) {
				filledSlots++;
				features.addBooleanFeature("slot" + s + "Filled",
						addFeaturesToDictionary);
			}
		}
		//numeric feature containing the number of filled slots.
		features.addNumericFeature("numFilledSlots", filledSlots,
				addFeaturesToDictionary);
		//add special features.
		addSpecialFeatures(features, addFeaturesToDictionary);
		return features;
	}

	/**
	 * This is to be implemented by a subclass to add domain-specific features
	 * to the mix.
	 *
	 * @param stdFeatures featureMap already filled with standard features.
	 * @param addFeaturesToDictionary this should be set to true during
	 * training, because then unknown features should be added to the
	 * FeatureDictionary. Outside training this should be set to false because
	 * then we want to treat every unknown feature as zero implicitly.
	 */
	public abstract void addSpecialFeatures(FeatureMap stdFeatures,
			boolean addFeaturesToDictionary);
	
	@Override
	public int hashCode() {
		int hash = 3;
		hash = 97 * hash + Objects.hashCode(this.core);
		hash = 97 * hash + Arrays.deepHashCode(this.slots);
		return hash;
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final RelationDataPoint<?> other = (RelationDataPoint<?>) obj;
		if (!Objects.equals(this.core, other.core)) {
			return false;
		}
		if (!Arrays.deepEquals(this.slots, other.slots)) {
			return false;
		}
		return true;
	}
	
	@Override
	public String toString() {
		final StringBuilder builder = new StringBuilder();
		builder.append("{\n");
		builder.append("core : ");
		builder.append(core.getCoveredText());
		builder.append("\n");
		for (final SlotCandidate slot : slots) {
			if (slot == null) {
				continue;
			}
			builder.append(slot.getAnnotation().getClass().getName());
			builder.append(" : ");
			builder.append(slot.getAnnotation().getCoveredText());
			builder.append("\n");
		}
		builder.append("}");
		return builder.toString();
	}

	/**
	 * Thre RelationTrainingDataReader is a highly complex TrainingDataReader to
	 * construct training DataPoints for a RelationClassifier. As mentioned in
	 * the RelationDataPoint documentation we want to use the classifier output
	 * for the core instance and the slot instances as basis for the
	 * RelationClassifier. Therefore it is necessary to train these classifiers
	 * before training a RelationClassifier and use the respective Classifiers
	 * as an argument for the constructor of the RelationTrainingDataReader. The
	 * order of classifiers is specified by the implementing subclasses.
	 *
	 * As mentioned in the RawRelation class the search space for
	 * RelationDataPoints is exponential in the number of slots, as it is
	 * defined as the Cartesian product of the candidate sets for each slot and
	 * the core candidate set.
	 *
	 * As in the CoreSlotTrainingDataReader this is not so much a problem for
	 * positive examples as those are hand-crafted but it gets a problem for
	 * negative training examples. Therefore we introduce two restrictions:
	 *
	 * <ul>
	 * <li>Only those instances are used for core and slot that have a positive
	 * classification result.</li>
	 * <li>If that is not enough we restrict the candidates for each slot to
	 * a hard limit of MAXNEGATIVESLOTCANDIDATES. In that case we do not take
	 * the first ones but sample the candidates according to a uniform
	 * probability distribution.</li>
	 * </ul>
	 *
	 * @param <C> core class.
	 */
	public static abstract class RelationTrainingDataReader<C extends Annotation> implements TrainingDataReader {
		
		public static final int MAX_SLOT_CANDIDATES = 5;
		
		private final String rawRelationClass;
		private final Class<C> coreClass;
		private final Classifier coreClassifier;
		private final ArrayList<SlotSpecification> slotSpecifications = new ArrayList<>();    /*
		 * This limits the number of annotations that is considered for each
		 * slot:
		 * We enforce that core and slot are at most a certain number of
		 * characters away from each other.
		 */
		
		public static final int DEFAULT_DIST_LIMIT = 1000;
		private int distLimit = DEFAULT_DIST_LIMIT;
		
		public RelationTrainingDataReader(Class<C> coreClass, Classifier coreClassifier,
				String rawRelationClass) {
			this.coreClass = coreClass;
			this.coreClassifier = coreClassifier;
			this.rawRelationClass = rawRelationClass;
		}
		
		public Class<C> getCoreClass() {
			return coreClass;
		}
		
		public Classifier getCoreClassifier() {
			return coreClassifier;
		}
		
		public String getRawRelationClass() {
			return rawRelationClass;
		}
		
		public ArrayList<SlotSpecification> getSlotSpecifications() {
			return slotSpecifications;
		}
		
		public int getDistLimit() {
			return distLimit;
		}
		
		public void setDistLimit(int distLimit) {
			this.distLimit = distLimit;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void readTrainingData(JCas jcas,
				Collection<RawRelation> rawRelationInstances,
				Collection<LabeledDataPoint> trainingDataCollection) {
			final int S = slotSpecifications.size();
			/*
			 * Store all core instances w.r.t. their ID
			 */
			final Collection<C> coreInstances = JCasUtil.select(jcas, coreClass);
			final HashMap<Long, C> coreIDMapping = new HashMap<>();
			for (final C coreInstance : coreInstances) {
				coreIDMapping.put(coreInstance.getAnnotationId(), coreInstance);
			}
			/*
			 * Store all instances for the slots w.r.t. their ID.
			 */
			final ArrayList<HashMap<Long, Annotation>> slotIDMappings = new ArrayList<>();
			for (int s = 0; s < S; s++) {
				final SlotSpecification spec = slotSpecifications.get(s);
				final Collection<? extends Annotation> slotInstances
						= JCasUtil.select(jcas, spec.getSlotClass());
				final HashMap<Long, Annotation> slotIDMapping = new HashMap<>();
				for (final Annotation slotInstance : slotInstances) {
					slotIDMapping.put(slotInstance.getAnnotationId(), slotInstance);
				}
				slotIDMappings.add(slotIDMapping);
			}
			
			final HashSet<RelationDataPoint<C>> positiveData = new HashSet<>();
			/*
			 * construct positive training examples.
			 */
			for (final RawRelation relationInstance : rawRelationInstances) {
				if (!relationInstance.relationClass.equals(rawRelationClass)) {
					continue;
				}
				/*
				 * Right now each raw relation might contain multiple candidates
				 * for core and slot. Ideally this can later be replaced
				 */
				//Construct the list of core instances for this relation instance first.
				final ArrayList<SlotCandidate<C>> coreCandidates = new ArrayList<>();
				final HashSet<Long> remainingIDs = new HashSet<>(relationInstance.ids);
				for (final Long id : relationInstance.ids) {
					final C coreAnno = coreIDMapping.get(id);
					if (coreAnno == null) {
						continue;
					}
					final ClassificationResult coreClassification
							= coreClassifier.classify(
									createCoreDataPoint(coreAnno, jcas));
					coreCandidates.add(
							new SlotCandidate<>(coreAnno, coreClassification));
					remainingIDs.remove(id);
				}
				if (coreCandidates.isEmpty()) {
					continue;
				}
				for (final SlotCandidate<C> coreCandidate : coreCandidates) {
					//Now construct for each slot the set of instances within this relation instance.
					final ArrayList<ArrayList<SlotCandidate<? extends Annotation>>> slotCandidatess = new ArrayList<>();
					for (int s = 0; s < S; s++) {
						final SlotSpecification<? extends Annotation> spec = slotSpecifications.get(s);
						final HashMap<Long, Annotation> slotIDMapping = slotIDMappings.get(s);
						final ArrayList<SlotCandidate<? extends Annotation>> slotCandidates = new ArrayList<>();
						slotCandidatess.add(slotCandidates);
						//store IDs for deletion that have been found.
						final Stack<Long> deletionIDs = new Stack<>();
						for (final Long id : remainingIDs) {
							final Annotation slotCandidate = slotIDMapping.get(id);
							if (slotCandidate == null) {
								continue;
							}
							final ClassificationResult slotClassificaton
									= spec.getSlotClassifier().classify(
											createCoreSlotCombinationDataPoint(
													spec,
													coreCandidate.getAnnotation(),
													slotCandidate, jcas));
							slotCandidates.add(
									new SlotCandidate(slotCandidate, slotClassificaton));
							deletionIDs.push(id);
						}
						//delete all IDs that have been found.
						while (!deletionIDs.empty()) {
							remainingIDs.remove(deletionIDs.pop());
						}
						//add null as candidate if there is no "real" candidate for our slot.
						if (slotCandidates.isEmpty()) {
							slotCandidates.add(null);
						}
					}
					//give a warning if an ID could not be resolved.
					for (final Long id : remainingIDs) {
						System.err.println("Warning! ID " + id + " could not be resolved!");
					}
					/*
					 * Now construct the cartesian product of the core and slot
					 * candidates.
					 */
					constructCartesianProduct(coreCandidate, slotCandidatess, jcas, new Yes<C>(), positiveData);
				}
			}
			//add the result to the output.
			for (final RelationDataPoint<C> posDataPoint : positiveData) {
				trainingDataCollection.add(new LabeledDataPoint(true, posDataPoint));
			}
			//get the NegativeAcceptor.
			final NegativeAcceptor neg = new NegativeAcceptor(positiveData);

			//Now we do not need the ID mappings anymore and can clear our RAM
			coreIDMapping.clear();
			slotIDMappings.clear();

            //construct negative training examples.
			//retrieve the set of viable core candidates first.
			final ArrayList<SlotCandidate<C>> coreCandidates = new ArrayList<>();
			for (final C coreAnno : coreInstances) {
				final ClassificationResult coreClassification
						= coreClassifier.classify(
								createCoreDataPoint(coreAnno, jcas));
				//we do not use cores that are not accepted by the core classifier.
				if (!coreClassification.getLabel()) {
					continue;
				}
				coreCandidates.add(new SlotCandidate<>(coreAnno, coreClassification));
			}
			final Random rand = new Random();
			//if there are two many potential cores, limit the number.
			while (coreCandidates.size() > MAX_SLOT_CANDIDATES) {
				final int r = rand.nextInt(coreCandidates.size());
				coreCandidates.remove(r);
			}
			
			for (final SlotCandidate<C> coreCandidate : coreCandidates) {
				//create the set of viable candidates for each slot given the core.
				final ArrayList<ArrayList<SlotCandidate<? extends Annotation>>> slotCandidatess = new ArrayList<>();
				for (int s = 0; s < S; s++) {
					final SlotSpecification spec = slotSpecifications.get(s);
					final ArrayList<SlotCandidate<? extends Annotation>> slotCandidates
							= getSlotCandidates(spec, coreCandidate.getAnnotation(), jcas);
					slotCandidatess.add(slotCandidates);
				}
				final HashSet<RelationDataPoint<C>> negativeData = new HashSet<>();
				//construct the cartesian product.
				constructCartesianProduct(coreCandidate, slotCandidatess, jcas,
						neg, negativeData);
				//add the result to the output.
				for (final RelationDataPoint<C> negDataPoint : negativeData) {
					trainingDataCollection.add(new LabeledDataPoint(false, negDataPoint));
				}
			}
		}
		
		private <S extends Annotation> ArrayList<SlotCandidate<S>> getSlotCandidates(
				SlotSpecification<S> spec, C coreAnno, JCas jcas) {
			final ArrayList<SlotCandidate<S>> slotCandidates = new ArrayList<>();
			//the empty canidate is always possible.
			slotCandidates.add(null);
			final int cStart = coreAnno.getBegin();
			final int cEnd = coreAnno.getEnd();
			for (final S slotAnno : JCasUtil.select(jcas, spec.getSlotClass())) {
				/*
				 * We do not want to consider annotations that are further away
				 * than the distance limit.
				 *
				 * This is not a nice heuristic, because in principle the
				 * classifiers should do this automatically. But practical
				 * experience has shown that they do not do so necessarily.
				 */
				final int dist;
				if (slotAnno.getEnd() < cStart) {
					dist = cStart - slotAnno.getEnd();
				} else if (slotAnno.getBegin() > cEnd) {
					dist = slotAnno.getBegin() - cEnd;
				} else {
					dist = 0;
				}
				if (dist > distLimit) {
					continue;
				}
				final CoreSlotCombinationDataPoint<C, ? extends Annotation> slotComb
						= createCoreSlotCombinationDataPoint(
								spec, coreAnno, slotAnno, jcas);
				final ClassificationResult slotClassification
						= spec.getSlotClassifier().classify(slotComb);
				/*
				 * we do not want to use slot fillers with negative
				 * classification results.
				 */
				if (slotClassification.getLabel()) {
					slotCandidates.add(
							new SlotCandidate<>(slotAnno, slotClassification));
				}
			}
			//if we have more slot candidates than we want, we reduce the number.
			final Random rand = new Random();
			while (slotCandidates.size() > MAX_SLOT_CANDIDATES) {
				final int r = rand.nextInt(slotCandidates.size());
				slotCandidates.remove(r);
				
			}
			return slotCandidates;
		}

		/**
		 * This is just an interface for internal purposes: After constructing
		 * the cartesian product we might still have reasons to get rid of some
		 * constructed relation instances.
		 *
		 * While constructing negative data points, mainly, we do not want to
		 * use data points that are already positive data points.
		 */
		private static interface RelationAcceptor<C extends Annotation> {
			
			public boolean accepts(RelationDataPoint<C> point);
		}
		
		private static class Yes<C extends Annotation> implements RelationAcceptor<C> {
			
			@Override
			public boolean accepts(RelationDataPoint point) {
				return true;
			}
			
		}
		
		private static class NegativeAcceptor<C extends Annotation> implements RelationAcceptor<C> {
			
			private final HashSet<RelationDataPoint<C>> positiveData;
			
			public NegativeAcceptor(HashSet<RelationDataPoint<C>> positiveData) {
				this.positiveData = positiveData;
			}
			
			@Override
			public boolean accepts(RelationDataPoint<C> point) {
				return !positiveData.contains(point);
			}
		}
		
		private void constructCartesianProduct(final SlotCandidate<C> coreCandidate,
				final ArrayList<ArrayList<SlotCandidate<? extends Annotation>>> slotCandidatess,
				final JCas jcas, final RelationAcceptor<C> acc,
				final Collection<RelationDataPoint<C>> out) {
			final int S = slotSpecifications.size();
			final Stack<LinkedList<Integer>> cartesianStack = new Stack<>();
			cartesianStack.push(new LinkedList<Integer>());
			while (!cartesianStack.empty()) {
				final LinkedList<Integer> currentProduct = cartesianStack.pop();
				if (currentProduct.size() == S) {
					/*
					 * if we have accumulated one candidate for each
					 * slot, we are finished
					 */
					final SlotCandidate[] slots = new SlotCandidate[S];
					//construct slot arrays
					for (int s = 0; s < S; s++) {
						final int candidateIdx = currentProduct.get(s);
						final ArrayList<SlotCandidate<? extends Annotation>> slotCandidates
								= slotCandidatess.get(s);
						slots[s] = slotCandidates.get(candidateIdx);
					}
					//now construct the datapoint.
					final RelationDataPoint<C> relDataPoint
							= createDataPoint(coreCandidate.getAnnotation(),
									coreCandidate.getResult(),
									slots, jcas);
					//check if we should use it.
					if (!acc.accepts(relDataPoint)) {
						continue;
					}
					//add it to the output.
					out.add(relDataPoint);
					continue;
				}
				/*
				 * If we are not finished yet we iterate through all candidates
				 * of the current slot and copy a new product instance for all
				 * of them.
				 */
				final ArrayList<SlotCandidate<? extends Annotation>> slotCandidates
						= slotCandidatess.get(currentProduct.size());
				boolean needsCopy = false;
				for (int c = 0; c < slotCandidates.size(); c++) {
					final LinkedList<Integer> newProduct;
					if (!needsCopy) {
						newProduct = currentProduct;
						needsCopy = true;
					} else {
						newProduct = (LinkedList<Integer>) currentProduct.clone();
						newProduct.pollLast();
					}
					newProduct.add(c);
					cartesianStack.push(newProduct);
				}
			}
		}

		/**
		 * This method should create a RelationDataPoint given a core instance,
		 * a core classification and Candidates for all slots (note that the
		 * candidates might be null).
		 *
		 * @param core a core instance.
		 * @param coreClassification the classification of that core instance
		 * including the classification confidence (the label will be true).
		 * @param slotCandidates the candidates for each slot. Note that each
		 * entry
		 * of that array might be null if the slot is not filled.
		 * @param jcas the current JCas object.
		 *
		 * @return a RelationDataPoint containing the given information.
		 */
		public abstract RelationDataPoint<C> createDataPoint(C core,
				ClassificationResult coreClassification,
				SlotCandidate[] slotCandidates,
				JCas jcas);

		/**
		 * This method should create a CoreDataPoint given a core instance.
		 *
		 * @param core a core instance.
		 * @param jcas the current JCas object.
		 *
		 * @return a CoreDataPoint containing the given information.
		 */
		public abstract CoreDataPoint<C> createCoreDataPoint(C core, JCas jcas);

		/**
		 * This method should create a CoreSlotCombination given a
		 * SlotSpecification, a core instance and a Slot instance.
		 *
		 * @param spec the current SlotSpecification.
		 * @param core the current Core instance.
		 * @param slot the current slot instance.
		 * @param jcas the current JCas object.
		 *
		 * @return a CoreSlotCombination containing the given data.
		 */
		public abstract CoreSlotCombinationDataPoint<C, ? extends Annotation> createCoreSlotCombinationDataPoint(
				final SlotSpecification spec, C core, Annotation slot, JCas jcas);
		
	}
	
}
